package org.eclipse.jdt.internal.compiler.apt.util;

import java.util.Iterator;
import java.util.Collections;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Set;
import java.util.Map;

public class ManyToMany<T1, T2>
{
    private final Map<T1, Set<T2>> _forward;
    private final Map<T2, Set<T1>> _reverse;
    private boolean _dirty;
    
    public ManyToMany() {
        this._forward = new HashMap();
        this._reverse = new HashMap();
        this._dirty = false;
    }
    
    public synchronized boolean clear() {
        boolean hadContent = !this._forward.isEmpty() || !this._reverse.isEmpty();
        this._reverse.clear();
        this._forward.clear();
        this._dirty |= hadContent;
        return hadContent;
    }
    
    public synchronized void clearDirtyBit() {
        this._dirty = false;
    }
    
    public synchronized boolean containsKey(T1 key) {
        return this._forward.containsKey(key);
    }
    
    public synchronized boolean containsKeyValuePair(T1 key, T2 value) {
        Set<T2> values = (Set)this._forward.get(key);
        return values != null && values.contains(value);
    }
    
    public synchronized boolean containsValue(T2 value) {
        return this._reverse.containsKey(value);
    }
    
    public synchronized Set<T1> getKeys(T2 value) {
        Set<T1> keys = (Set)this._reverse.get(value);
        if (keys == null) {
            return Collections.emptySet();
        }
        return new HashSet(keys);
    }
    
    public synchronized Set<T2> getValues(T1 key) {
        Set<T2> values = (Set)this._forward.get(key);
        if (values == null) {
            return Collections.emptySet();
        }
        return new HashSet(values);
    }
    
    public synchronized Set<T1> getKeySet() {
        Set<T1> keys = new HashSet(this._forward.keySet());
        return keys;
    }
    
    public synchronized Set<T2> getValueSet() {
        Set<T2> values = new HashSet(this._reverse.keySet());
        return values;
    }
    
    public synchronized boolean isDirty() {
        return this._dirty;
    }
    
    public synchronized boolean keyHasOtherValues(T1 key, T2 value) {
        Set<T2> values = (Set)this._forward.get(key);
        if (values == null) {
            return false;
        }
        int size = values.size();
        return size != 0 && (size > 1 || !values.contains(value));
    }
    
    public synchronized boolean put(T1 key, T2 value) {
        Set<T2> values = (Set)this._forward.get(key);
        if (values == null) {
            values = new HashSet();
            this._forward.put(key, values);
        }
        boolean added = values.add(value);
        this._dirty |= added;
        Set<T1> keys = (Set)this._reverse.get(value);
        if (keys == null) {
            keys = new HashSet();
            this._reverse.put(value, keys);
        }
        keys.add(key);
        assert !(!this.checkIntegrity());
        return added;
    }
    
    public synchronized boolean remove(T1 key, T2 value) {
        Set<T2> values = (Set)this._forward.get(key);
        if (values == null) {
            assert !(!this.checkIntegrity());
            return false;
        }
        else {
            boolean removed = values.remove(value);
            if (values.isEmpty()) {
                this._forward.remove(key);
            }
            if (removed) {
                this._dirty = true;
                Set<T1> keys = (Set)this._reverse.get(value);
                keys.remove(key);
                if (keys.isEmpty()) {
                    this._reverse.remove(value);
                }
            }
            assert !(!this.checkIntegrity());
            return removed;
        }
    }
    
    public synchronized boolean removeKey(T1 key) {
        Set<T2> values = (Set)this._forward.get(key);
        if (values == null) {
            assert !(!this.checkIntegrity());
            return false;
        }
        else {
            for (T2 value : values) {
                Set<T1> keys = (Set)this._reverse.get(value);
                if (keys != null) {
                    keys.remove(key);
                    if (keys.isEmpty()) {
                        this._reverse.remove(value);
                    }
                    else {
                        continue;
                    }
                }
                else {
                    continue;
                }
            }
            this._forward.remove(key);
            this._dirty = true;
            assert !(!this.checkIntegrity());
            return true;
        }
    }
    
    public synchronized boolean removeValue(T2 value) {
        Set<T1> keys = (Set)this._reverse.get(value);
        if (keys == null) {
            assert !(!this.checkIntegrity());
            return false;
        }
        else {
            for (T1 key : keys) {
                Set<T2> values = (Set)this._forward.get(key);
                if (values != null) {
                    values.remove(value);
                    if (values.isEmpty()) {
                        this._forward.remove(key);
                    }
                    else {
                        continue;
                    }
                }
                else {
                    continue;
                }
            }
            this._reverse.remove(value);
            this._dirty = true;
            assert !(!this.checkIntegrity());
            return true;
        }
    }
    
    public synchronized boolean valueHasOtherKeys(T2 value, T1 key) {
        Set<T1> keys = (Set)this._reverse.get(key);
        if (keys == null) {
            return false;
        }
        int size = keys.size();
        return size != 0 && (size > 1 || !keys.contains(key));
    }
    
    private boolean checkIntegrity() {
        for (Map.Entry<T1, Set<T2>> entry : this._forward.entrySet()) {
            Set<T2> values = (Set)entry.getValue();
            if (values.isEmpty()) {
                throw new IllegalStateException("Integrity compromised: forward map contains an empty set");
            }
            for (T2 value : values) {
                Set<T1> keys = (Set)this._reverse.get(value);
                if (keys == null || !keys.contains(entry.getKey())) {
                    throw new IllegalStateException("Integrity compromised: forward map contains an entry missing from reverse map: " + value);
                }
                continue;
            }
        }
        for (Map.Entry<T2, Set<T1>> entry2 : this._reverse.entrySet()) {
            Set<T1> keys2 = (Set)entry2.getValue();
            if (keys2.isEmpty()) {
                throw new IllegalStateException("Integrity compromised: reverse map contains an empty set");
            }
            for (T1 key : keys2) {
                Set<T2> values2 = (Set)this._forward.get(key);
                if (values2 == null || !values2.contains(entry2.getKey())) {
                    throw new IllegalStateException("Integrity compromised: reverse map contains an entry missing from forward map: " + key);
                }
                continue;
            }
        }
        return true;
    }
}
